/*
  This code is released under the Antivirial License.

  http://www.ecstaticlyrics.com/antiviral/

  The only condition of the license is that you may not change the license
  terms.  Because of this, you may *not* include this code in GPL-licensed
  projects, as the GPL requires that the entire project be GPL-licensed.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <complex.h>

#include <linux/videodev2.h>
#include <sys/types.h>  // for open()
#include <sys/stat.h>   // for open()
#include <fcntl.h>      // for open()
#include <sys/ioctl.h>  // for ioctl()
#include <unistd.h>     // for read()

#include "easy.h"
#include "thread.h"
#include "buffer.h"
#include "fft.h"
#include "user.h"

static volatile int thread_run_flag = 0;
static volatile int thread_exit_flag = 0;
static volatile int initialized = 0;

static int fd;

MUTEX frequency_lock;
static int frequency_get;
static int frequency_set;

MUTEX sample_rate_lock;
static int sample_rate_get;
static int sample_rate_set;

static int buffer_size;

void open_sdr() {

  char *pathname = "/dev/swradio0";

  fd = open(pathname, O_RDWR);
  if (fd == -1) {
    fprintf(stderr, "ERROR: Failed to open %s\n", pathname);
    exit(1);
  };
  fprintf(stderr, "\nSuccessfully opened %s\n", pathname);

  int result;

  // Query capabilities...

  struct v4l2_capability info;
  result = ioctl(fd, VIDIOC_QUERYCAP, &info);
  if (result < 0) {
    fprintf(stderr, "ERROR: Driver does not support VIDIOC_QUERYCAP, which is quite strange.\n");
    exit(1);
  };

  fprintf(stderr, "\nVIDIOC_QUERYCAP results:\n");
  fprintf(stderr, "  driver: %s\n", info.driver);
  fprintf(stderr, "  card: %s\n", info.card);
  fprintf(stderr, "  bus_info: %s\n", info.bus_info);
  fprintf(stderr, "  version: %d.%d.%d\n", (info.version >> 16), (info.version >> 8) & 255, info.version & 255);

  int caps = info.capabilities;
  fprintf(stderr, "  capabilities: %#010x\n", caps);
  if (caps & V4L2_CAP_DEVICE_CAPS) {
    caps = info.device_caps;
    fprintf(stderr, "  device_caps: %#010x\n", caps);
  };

  if (caps != 0x05310000) {
    fprintf(stderr, "\nNOTE:  The device_caps field doesn't match what my driver returns.\n");
    fprintf(stderr, "However, I don't know what's truly necessary, so I'll ignore it.\n");
  };

/*
  Flag meanings:

  0x00010000 Has a tuner.
  0x00100000 Supports SDR capture.
  0x00200000 Supports v4l2_pix_format extended fields.
  0x01000000 Supports read() and/or write() I/O methods.
  0x04000000 Supports streaming I/O method.
*/

  // Query controls...

/*
  Available controls:
    10616833: RF Tuner Controls, min=0, max=0, step=0, default=0, flags=0x00000044
    10619147: Bandwidth, Auto, min=0, max=1, step=1, default=1, flags=0x00000008, value=1
    10619148: Bandwidth, min=0, max=8000000, step=100000, default=0, flags=0x00000030, value=1000000
*/

  struct v4l2_queryctrl qc;

  fprintf(stderr, "\nAvailable controls:\n");
  for (int i = 0; ; ) {
    memset(&qc, 0, sizeof(qc));
    qc.id = i | V4L2_CTRL_FLAG_NEXT_CTRL;
    result = ioctl(fd, VIDIOC_QUERYCTRL, &qc);
    if (result < 0) break;
    i = qc.id;
    struct v4l2_control c;
    memset(&c, 0, sizeof(c));
    c.id = qc.id;
    result = ioctl(fd, VIDIOC_G_CTRL, &c);
    if (result < 0) {
      fprintf(stderr, "  %i: %s, min=%d, max=%d, step=%d, default=%d, flags=%#010x\n", i, qc.name, qc.minimum, qc.maximum, qc.step, qc.default_value, qc.flags);
    } else {
      fprintf(stderr, "  %i: %s, min=%d, max=%d, step=%d, default=%d, flags=%#010x, value=%d\n", i, qc.name, qc.minimum, qc.maximum, qc.step, qc.default_value, qc.flags, c.value);
    };
  };

  // Enumerate data formats...

/*
Available data formats:
  0: Complex U8
*/

  fprintf(stderr, "\nAvailable data formats:\n");
  struct v4l2_fmtdesc fm;
  for (int i = 0; ; i++) {
    memset(&fm, 0, sizeof(fm));
    fm.index = i;
    fm.type = V4L2_BUF_TYPE_SDR_CAPTURE;
    result = ioctl(fd, VIDIOC_ENUM_FMT, &fm);
    if (result < 0) break;
    fprintf(stderr, "  %d: %s, pixelformat=%#010x\n", i, fm.description, fm.pixelformat);
  };

  // Choose the Complex U8 format...

  buffer_size = 65536;

  #if 0

  struct v4l2_format fuck;
  memset(&fuck, 0, sizeof(fuck));
  fuck.type = V4L2_BUF_TYPE_SDR_CAPTURE;
  fuck.fmt.sdr.pixelformat = V4L2_SDR_FMT_CU8;
  result = ioctl(fd, VIDIOC_S_FMT, &fuck);
  if (result < 0) {
    fprintf(stderr, "ERROR: Failed to set capture format.");
    exit(1);
  };
  fprintf(stderr, "\nfmt.sdr.buffersize=%d\n", fuck.fmt.sdr.buffersize);
  buffer_size = fuck.fmt.sdr.buffersize;

  #endif

  // Query tuners...

/*
  Available tuners:
  0: name=ADC: Realtek RTL2832, type=4, capability=0x1400, rangelow=300000, rangehigh=3200000, frequency=2400000 (2.400 MHz)
  1: name=RF: <unknown>, type=5, capability=0x1400, rangelow=50000000, rangehigh=2000000000, frequency=454000000 (454.000 MHz)
*/

  fprintf(stderr, "\nAvailable tuners:\n");
  struct v4l2_tuner t;
  for (int i = 0; ; i++) {
    memset(&t, 0, sizeof(t));
    t.index = i;
    result = ioctl(fd, VIDIOC_G_TUNER, &t);
    if (result < 0) break;
    struct v4l2_frequency f;
    memset(&f, 0, sizeof(f));
    f.tuner = t.index;
    f.type = t.type;
    if (t.index == 0) f.frequency = 2400000;
    if (t.index == 1) f.frequency = 139000000;
    result = ioctl(fd, VIDIOC_S_FREQUENCY, &f);
    result = ioctl(fd, VIDIOC_G_FREQUENCY, &f);
    if (result < 0) {
      fprintf(stderr, "  %d: name=%s, type=%d, capability=%#06x, rangelow=%d, rangehigh=%d\n", i, t.name, t.type, t.capability, t.rangelow, t.rangehigh);
    } else {
      fprintf(stderr, "  %d: name=%s, type=%d, capability=%#06x, rangelow=%d, rangehigh=%d, frequency=%d (%0.3f MHz)\n", i, t.name, t.type, t.capability, t.rangelow, t.rangehigh, f.frequency, f.frequency / 1000000.0);
    };
    if (t.index == 0) sample_rate_get = sample_rate_set = f.frequency;
    if (t.index == 1) frequency_get = frequency_set = f.frequency;
    if (t.capability & V4L2_TUNER_CAP_FREQ_BANDS) {
      struct v4l2_frequency_band fb;
      for (int j = 0; ; j++) {
        memset(&fb, 0, sizeof(fb));
        fb.tuner = t.index;
        fb.type = t.type;
        fb.index = j;
        result = ioctl(fd, VIDIOC_ENUM_FREQ_BANDS);
        //if (result < 0) fprintf(stderr, "error: %s\n", error_string(errno));
        if (result < 0) break;
        fprintf(stderr, "        %d: capability=%#06x, rangelow=%d, rangehigh=%d, modulation=%d\n", j, fb.capability, fb.rangelow, fb.rangehigh, fb.modulation);
      };
    };

    #if 0

    // Determine valid range of frequencies...

    // ADC rate: 300000, 900001-2800000, 3200000
    // Demodulation: 50000000-2000000000

    fprintf(stderr, "        Valid Ranges: ");
    int first = -1; int step = 1;
    if (t.index == 1) step = 1000;
    int j;
    for (j = t.rangelow; j <= t.rangehigh; j += step) {
      struct v4l2_frequency f;
      memset(&f, 0, sizeof(f));
      f.tuner = t.index;
      f.type = t.type;
      f.frequency = j;
      result = ioctl(fd, VIDIOC_S_FREQUENCY, &f);
      if (result < 0) continue;
      f.frequency = 12345678;
      result = ioctl(fd, VIDIOC_G_FREQUENCY, &f);
      if (result < 0) continue;
      if (j == f.frequency) {
        if (first == -1) {
          fprintf(stderr, "%d", j);
          first = j;
        };
      } else {
        if (first != -1) {
          if (first == j - step) {
            fprintf(stderr, ", ");
          } else {
            fprintf(stderr, "-%d, ", j - step);
          };
        };
        first = -1;
      };
    };
    if (first != -1) {
      if (first != j - step) {
        fprintf(stderr, "-%d", j - step);
      };
    };
    fprintf(stderr, "\n");
    first = -1;

    #endif

  };

  fprintf(stderr, "\n");

};

static float complex filter_2400_lowpass_100 (float complex input) {
  #define NZEROS 7
  #define NPOLES 7
  #define GAIN   2.619173119e+06

  static float complex xv[NZEROS+1] = {}, yv[NPOLES+1] = {};

  xv[0] = xv[1]; xv[1] = xv[2]; xv[2] = xv[3]; xv[3] = xv[4]; xv[4] = xv[5]; xv[5] = xv[6]; xv[6] = xv[7];
  xv[7] = input / GAIN;
  yv[0] = yv[1]; yv[1] = yv[2]; yv[2] = yv[3]; yv[3] = yv[4]; yv[4] = yv[5]; yv[5] = yv[6]; yv[6] = yv[7];
  yv[7] =   (xv[0] + xv[7]) + 7 * (xv[1] + xv[6]) + 21 * (xv[2] + xv[5])
               + 35 * (xv[3] + xv[4])
               + (  0.3067227344 * yv[0]) + ( -2.5080683796 * yv[1])
               + (  8.8240801929 * yv[2]) + (-17.3204029220 * yv[3])
               + ( 20.4909963760 * yv[4]) + (-14.6163880230 * yv[5])
               + (  5.8230111503 * yv[6]);
  return yv[7];

  #undef NZEROS
  #undef NPOLES
  #undef GAIN
};

static float filter_100_lowpass_15 (float input) {
  #define NZEROS 7
  #define NPOLES 7
  #define GAIN   1.038535862e+03

  static float xv[NZEROS+1], yv[NPOLES+1];

  xv[0] = xv[1]; xv[1] = xv[2]; xv[2] = xv[3]; xv[3] = xv[4]; xv[4] = xv[5]; xv[5] = xv[6]; xv[6] = xv[7];
  xv[7] = input / GAIN;
  yv[0] = yv[1]; yv[1] = yv[2]; yv[2] = yv[3]; yv[3] = yv[4]; yv[4] = yv[5]; yv[5] = yv[6]; yv[6] = yv[7];
  yv[7] =   (xv[0] + xv[7]) + 7 * (xv[1] + xv[6]) + 21 * (xv[2] + xv[5])
               + 35 * (xv[3] + xv[4])
               + (  0.0116627280 * yv[0]) + ( -0.1308524516 * yv[1])
               + (  0.6509456985 * yv[2]) + ( -1.8758627160 * yv[3])
               + (  3.4051503945 * yv[4]) + ( -3.9668078920 * yv[5])
               + (  2.7825138004 * yv[6]);
  return yv[7];

  #undef NZEROS
  #undef NPOLES
  #undef GAIN
};

static float complex filter_100_lowpass_5 (float complex input, int n) {
  #define NZEROS 7
  #define NPOLES 7
  #define GAIN   8.132396142e+05

  static float complex xv[16][NZEROS+1] = {}, yv[16][NPOLES+1] = {};

  xv[n][0] = xv[n][1]; xv[n][1] = xv[n][2]; xv[n][2] = xv[n][3]; xv[n][3] = xv[n][4]; xv[n][4] = xv[n][5]; xv[n][5] = xv[n][6]; xv[n][6] = xv[n][7];
  xv[n][7] = input / GAIN;
  yv[n][0] = yv[n][1]; yv[n][1] = yv[n][2]; yv[n][2] = yv[n][3]; yv[n][3] = yv[n][4]; yv[n][4] = yv[n][5]; yv[n][5] = yv[n][6]; yv[n][6] = yv[n][7];
  yv[n][7] =   (xv[n][0] + xv[n][7]) + 7 * (xv[n][1] + xv[n][6]) + 21 * (xv[n][2] + xv[n][5])
               + 35 * (xv[n][3] + xv[n][4])
               + (  0.2418731990 * yv[n][0]) + ( -2.0343997965 * yv[n][1])
               + (  7.3738011769 * yv[n][2]) + (-14.9365219320 * yv[n][3])
               + ( 18.2708525180 * yv[n][4]) + (-13.5047447000 * yv[n][5])
               + (  5.5889821399 * yv[n][6]);
  return yv[n][7];

  #undef NZEROS
  #undef NPOLES
  #undef GAIN
};

float v4l2_offset_frequency = 0.199f;

static void *thread (void *parameter) {

  open_sdr();

  MUTEX_INIT(frequency_lock);
  MUTEX_INIT(sample_rate_lock);

  initialized = 1;

  #define FFT_SIZE 1024
  float complex input[FFT_SIZE];
  float complex output[FFT_SIZE];
  struct fft_state fft_state = {};
  fft_init(&fft_state, FFT_SIZE, 0);

  BUFFER buffer = {};

  if (buffer_size > 2 * FFT_SIZE) buffer_size = 2 * FFT_SIZE;

  int every_n = 0;

  while (!thread_exit_flag) {

    char *pointer = buffer_allocate(buffer, buffer_size);
    //printf("read(%d, %p, %d)\n", fd, pointer, buffer_size);
    int count = read(fd, pointer, buffer_size);
    if (count <= 0) {
      fprintf(stderr, "ERROR: read() from SDR returned %d\n", count);
      exit(1);
    };
    #if 0
    for (int i = 0; i < count; i++) {
      buffer.pointer[buffer_size + i] = easy_random(256);
    };
    #endif
    buffer.size += count;

    while (buffer.size > 2 * FFT_SIZE) {

      if (every_n++ == 0) {
        for (int i = 0; i < FFT_SIZE; i++) {
          input[i] = ((buffer.pointer[2 * i] - 127.5f) + (buffer.pointer[2 * i + 1] - 127.5f) * I) / 127.5f;
          #if 0
          static float a = 0.0f;
          a += 2 * M_PI * v4l2_offset_frequency / 2400000.0f;
          if (a > M_PI) a -= 2 * M_PI;
          input[i] *= cexpf(-a * I);
          #endif
        };
        fft_fast(&fft_state, output, input);
        float values[FFT_SIZE];
        for (int i = 0; i < FFT_SIZE; i++) {
          values[i] = cabsf(output[(i + (FFT_SIZE >> 1)) & (FFT_SIZE - 1)]);
        };
        user_waterfall_append(values);
      };
      if (every_n == 16) every_n = 0;

      for (int i = 0; i < FFT_SIZE; i++) {

        float complex sample = ((buffer.pointer[2 * i] - 127.5f) + (buffer.pointer[2 * i + 1] - 127.5f) * I) / 127.5f;

        static float a = 0.0f;
        a += 2 * M_PI * v4l2_offset_frequency / 2400000.0f;
        if (a > M_PI) a -= 2 * M_PI;
        sample *= cexpf(-a * I);

        // ...demodulate FM here...

        #define WIDE 0
        #define FM 0
        #define AM 0
        #define LSB 0
        #define USB 1

        sample = filter_2400_lowpass_100 (sample);

        #if WIDE & FM
          static float complex phase = 1.0;
          float complex change = sample / phase;
          float audio = atan2f(cimagf(change), crealf(change)) / M_PI;
          if (cabsf(sample) > 0.001f) {
            phase = sample;
          };
        #endif

        static int every_24 = 0;
        if (every_24++ == 0) {

          #if !WIDE
            sample = filter_100_lowpass_5 (sample, 0);
          #endif

          #if AM
            float audio = cabsf(sample);
          #endif

          #if LSB | USB
            static float complex a = 0.0f;
            a += 2 * M_PI * I * 5.0f / 100.0f;
            #if LSB
            sample *= cexpf(-a);
            #else
            sample *= cexpf(+a);
            #endif
            sample = filter_100_lowpass_5 (sample, 1);
            #if LSB
            sample *= cexpf(+a);
            #else
            sample *= cexpf(-a);
            #endif
            float audio = crealf(sample);
          #endif

          #if FM
            static float complex phase = 1.0;
            float complex change = sample / phase;
            float audio = atan2f(cimagf(change), crealf(change)) / M_PI;
            if (cabsf(sample) > 0.001f) {
              phase = sample;
            };
            audio = filter_100_lowpass_15 (audio);
          #endif

          static int every_5 = 0;
          if (every_5++ == 0) {
            write(1, &audio, 4);
            //fprintf(stderr, "audio = %f\n", audio);
            //static int counter = 0; if (counter++ > 5) exit(1);
          };
          if (every_5 >= 5) every_5 = 0;

        };
        if (every_24 >= 24) every_24 = 0;

      };

      buffer_free(buffer, 2 * FFT_SIZE);

    };

    /*
      Available controls:
        10616833: RF Tuner Controls, min=0, max=0, step=0, default=0, flags=0x00000044
        10619147: Bandwidth, Auto, min=0, max=1, step=1, default=1, flags=0x00000008, value=1
        10619148: Bandwidth, min=0, max=8000000, step=100000, default=0, flags=0x00000030, value=1000000
    */

    static int once = 1;
    if (once) {
      once = 0;
      struct v4l2_control control;
      control.id = 10619147;
      //control.id = 10619148;
      control.value = 0;
      fprintf(stderr, "calling ioctl()...\n");
      int result = ioctl(fd, VIDIOC_S_CTRL, &control);
      fprintf(stderr, "returned %d\n", result);
    };

    static int every_c = 0;
    if (every_c++ == 0) {

      #if 0
      struct v4l2_frequency f;
      memset(&f, 0, sizeof(f));
      f.tuner = 1;
      f.type = 5;
      static int cv;
      f.frequency = 454000000 + cv;
      fprintf(stderr, "calling ioctl()...\n");
      int result = ioctl(fd, VIDIOC_S_FREQUENCY, &f);
      fprintf(stderr, "returned %d\n", result);
      cv += 100000;

      static int cv = 0;
      struct v4l2_control control;
      //control.id = 10619147;
      control.id = 10619148;
      control.value = cv;
      fprintf(stderr, "calling ioctl()...\n");
      int result = ioctl(fd, VIDIOC_S_CTRL, &control);
      if (result < 0) {
        fprintf(stderr, "failed to set control\n");
      } else {
        fprintf(stderr, "control is now %d\n", cv);
      }
      cv += 8000000;
      if (cv > 8000000) cv = 0;
      #endif

    };
    if (every_c == 10000) every_c = 0;



  };

  fft_free(&fft_state);
  buffer_kill(buffer);

  thread_run_flag = 0;

  thread_exit();

};

void v4l2_initialize () {

  thread_run_flag = 1;
  thread_exit_flag = 0;
  thread_create(thread, NULL);

  while (!initialized) easy_sleep(0.01);

};

void v4l2_terminate () {

  if (thread_run_flag) {
    thread_exit_flag = 1;
    while (thread_run_flag) easy_sleep(0.01);
  };

};
